/*
  medx_anova - fMRI ANOVA analysis 

  Stuart Clare, FMRIB Physics Group

  Copyright (c) 1998-2000 University of Oxford
   
  Clare, S. et.al. (1999) Magnetic Resonance in Medicine. 42,1117-1122.
*/

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "hfunc.h"
#include "ftoz.h"

/* usage: medx_anova infile Ffile Zfile %thold offset points [-av file] */

#define MAXLEN 128

int main(int argc,char *argv[])
{
  char infile[MAXLEN],foutfile[MAXLEN],zoutfile[MAXLEN],avfile[MAXLEN];
  short *image,thold;
  short max=0;
  double *F,*Z,thresh;
  int i,j,k,l,points,cycles,offset;
  int ppl,lpi,ipv,vols,type,av=0;
  long volsize,c;
  float vx,vy,vz,*fdata;
  double *data,T,C,SST,SSTr,Ti,SSE;
  FILE *ifp,*fofp,*zofp,*afp;

  if(argc<7){
    printf("usage: medx_anova infile Ffile Zfile %%thold offset points [-av file]\n");
    return(1);
  }

  sprintf(infile,"%s.img",argv[1]);
  sprintf(foutfile,"%s.img",argv[2]);
  sprintf(zoutfile,"%s.img",argv[3]);
  sscanf(argv[4],"%lf",&thresh);
  sscanf(argv[5],"%d",&offset);
  sscanf(argv[6],"%d",&points);
  if(argc>7){
    if(!strcmp(argv[7],"-av")){
      av=1;
      sprintf(avfile,"%s.img",argv[8]);
    }
  }

  /* Read header */
  if(hread_plus(infile,&ppl,&lpi,&ipv,&vols,&type,&vx,&vy,&vz)==-1){
    printf("Cannot read header file: %s\b\b\b\b.hdr\n",infile);
    return(1);
  }
  if(type!=DT_SIGNED_SHORT){
    printf("Data type must be signed short");
    return(1);
  }
  volsize=ppl*lpi*ipv;
  cycles=(vols-offset)/points;
  vols=(cycles*points);
	
  /* Open files and malloc */
  if((ifp=fopen(infile,"r"))==NULL){
    printf("Can't open : %s\n",infile);
    return(1);
  }
  if(volsize>ppl*vols){
    if((image=(short *)malloc(volsize*sizeof(short)))==NULL){
      printf("Malloc failed.\n");
      return(1);
    }
  } else {
    if((image=(short *)malloc(ppl*vols*sizeof(short)))==NULL){
      printf("Malloc failed.\n");
      return(1);
    }
  }
  if((data=(double *)malloc(vols*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((F=(double *)malloc(volsize*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((Z=(double *)malloc(volsize*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((fdata=(float *)malloc(volsize*sizeof(float)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }

  /* Read in first image and find max */
  fseek(ifp,0,SEEK_SET);
  if((fread(image,sizeof(short),volsize,ifp))!=volsize){
    printf("Read error.\n");
    return(1);
  }
  for(i=0;i<volsize;i++)if(image[i]>max)max=image[i];
  thold=max*thresh/100;

  /* For each line in image */
  for(i=0;i<ipv*lpi;i++){	
    /* Read in data */	
    for(k=0;k<vols;k++){
      fseek(ifp,(((k+offset)*volsize)+(i*ppl))*sizeof(short),0);
      if((fread(&image[k*ppl],sizeof(short),ppl,ifp))!=ppl){
	printf("Read error.\n");
	return(1);
      }
    }
    /* For each point in line */
    for(j=0;j<ppl;j++){		
      if(image[j]>thold){
	/* Unaveraged data set */
	for(k=0;k<vols;k++)data[k]=(double)image[(k*ppl)+j];

	T=SST=0.0;
	for(k=0;k<vols;k++){
	  T+=data[k];
	  SST+=(data[k]*data[k]);
	}
	C=(T*T)/vols;
	SST-=C;
	SSTr=0.0;
	for(k=0;k<points;k++){
	  for(l=0,Ti=0.0;l<cycles;l++){
	    Ti+=data[(l*points)+k];
	  }
	  SSTr+=(Ti*Ti);
	}
	SSTr/=cycles;
	SSTr-=C;
	SSE=SST-SSTr;
	F[(i*ppl)+j]=(SSTr/(points-1))/(SSE/(points*(cycles-1)));
      }
      else F[(i*ppl)+j]=0;
    }
  }

  /* Create average image */
  if(av){
    if((afp=fopen(avfile,"w+"))==NULL){
      printf("Can't open file: %s\n",avfile);
      return(1);
    }
    for(k=0;k<points;k++){
      for(c=0;c<volsize;c++)fdata[c]=0.0;
      for(j=0;j<cycles;j++){
	fseek(ifp,(j*points+k+offset)*volsize*sizeof(short),SEEK_SET);
	if((fread(image,sizeof(short),volsize,ifp))!=volsize){
	  printf("Read error.\n");
	  return(1);
	}
	for(c=0;c<volsize;c++)fdata[c]+=((float)image[c]/(float)cycles);
      }
      fseek(afp,k*volsize*sizeof(float),SEEK_SET);
      if(fwrite(fdata,sizeof(float),volsize,afp)!=volsize){
	printf("Write error\n");
	return(1);
      }
    }
    if(hmake_plus(avfile,ppl,lpi,ipv,points,DT_FLOAT,max,0,vx,vy,vz)==-1){
      printf("Cannot create header: %s\b\b\b\b.hdr\n",foutfile);
      return(1);
    }
    fclose(afp);
  }
  fclose(ifp);

  /* Transform F to Z */
  ftoz(F,Z,volsize,points-1,points*(cycles-1));

  /* Open output files */
  if((fofp=fopen(foutfile,"w"))==NULL){
    printf("Can't open file: %s\n",foutfile);
    return(1);
  }
  if((zofp=fopen(zoutfile,"w"))==NULL){
    printf("Can't open file: %s\n",zoutfile);
    return(1);
  }

  /* Write statistics and create header */
  for(c=0;c<volsize;c++)fdata[c]=(float)F[c];
  fseek(fofp,0,0);
  if(fwrite(fdata,sizeof(float),volsize,fofp)!=volsize){
    printf("Write error\n");
    return(1);
  }	
  if(hmake_plus(foutfile,ppl,lpi,ipv,1,DT_FLOAT,10,0,vx,vy,vz)==-1){
    printf("Cannot create header: %s\b\b\b\b.hdr\n",foutfile);
    return(1);
  }

  for(c=0;c<volsize;c++)fdata[c]=(float)Z[c];
  fseek(zofp,0,0);
  if(fwrite(fdata,sizeof(float),volsize,zofp)!=volsize){
    printf("Write error\n");
    return(1);
  }	
  if(hmake_plus(zoutfile,ppl,lpi,ipv,1,DT_FLOAT,10,0,vx,vy,vz)==-1){
    printf("Cannot create header: %s\b\b\b\b.hdr\n",zoutfile);
    return(1);
  }
  fclose(zofp);
  fclose(fofp);

  return(0);
}
