/* {{{ Copyright etc. */

/*  slicer - produce any slicings of a 3D image

    Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

/* }}} */
/* {{{ defines, includes and typedefs */

#include "libss/libss.h"
#include "libss/libavw.h"
#include "libss/libpicio.h"

void usage(void);

int x_size, y_size, z_size, x_size_pic, y_size_pic, z_size_pic;

/* }}} */
/* {{{ usage */

void usage(void)
{
  fprintf(stderr,"\nUsage: slicer <fileroot> [secondfileroot] [-s <scale>] [-i <intmin> <intmax>] [output options - any number of these]\nNote: the -s and -i options, if used, must come in the specified order\n\nOutput options:\n");
  fprintf(stderr,"[-a <filename>] : output mid-sagittal, -coronal and -axial slices into one image\n");
  fprintf(stderr,"[-x/y/z <slice> <filename>] : output sagittal, coronal or axial slices -\n  if slice is positive it is a fraction of image dimension, if negative, an absolute slice number\n");
  exit(1);
}

/* }}} */
/* {{{ make slice images */

/* {{{ write sagittal slice */

void sag(float xx, image_struct im, image_struct im2, int compare, unsigned char *pic, int width)
{
  float yy, zz;
  int   tmp, y, z;
  FDT   interp;

  if (xx<0)
    xx=-xx;
  else
    xx*=(float)x_size;
  if (xx>=x_size-1) xx=x_size-1.0001;

  for(y=0; y<y_size_pic; y++)
    for(z=0; z<z_size_pic; z++)
      {
	yy=((double)y)/im.yv;
	zz=((double)z)/im.zv;

	if (yy>=y_size-1) yy=y_size-1.0001;
	if (zz>=z_size-1) zz=z_size-1.0001;

	interp = TLI(im,xx,yy,zz);
	tmp = ( ( ((double)interp) - ((double)im.thresh2) ) * 254.0 /
		( ((double)im.thresh98) - ((double)im.thresh2) ) );
	tmp = MIN(254,MAX(0,tmp));

	if ( (compare) &&
	     (IA(im2.i,(int)(xx+0.5),(int)(yy+0.5),(int)(zz+0.5))>0) &&
	     ((y+z)%2>-10) )
	  tmp=255;

	pic[(z_size_pic-1-z)*width+y]= (unsigned char)tmp;
      }
}

/* }}} */
/* {{{ write coronal slice */

void cor(float yy, image_struct im, image_struct im2, int compare, unsigned char *pic, int width)
{
float xx, zz;
int   tmp, x, z;
FDT   interp;

  if (yy<0)
    yy=-yy;
  else
    yy*=(float)y_size;
  if (yy>=y_size-1) yy=y_size-1.0001;

  for(x=0; x<x_size_pic; x++)
    for(z=0; z<z_size_pic; z++)
      {
	xx=((double)x)/im.xv;
	zz=((double)z)/im.zv;

	if (xx>=x_size-1) xx=x_size-1.0001;
	if (zz>=z_size-1) zz=z_size-1.0001;

	interp = TLI(im,xx,yy,zz);
	tmp = ( ( ((double)interp) - ((double)im.thresh2) ) * 254.0 /
		( ((double)im.thresh98) - ((double)im.thresh2) ) );
	tmp = MIN(254,MAX(0,tmp));

	if ( (compare) &&
	     (IA(im2.i,(int)(xx+0.5),(int)(yy+0.5),(int)(zz+0.5))>0) &&
	     ((x+z)%2>-10) )
	  tmp=255;

	pic[(z_size_pic-1-z)*width+x]= (unsigned char)tmp;
      }
}

/* }}} */
/* {{{ write axial slice */

void axi(float zz, image_struct im, image_struct im2, int compare, unsigned char *pic, int width)
{
float xx, yy;
int   tmp, x, y;
FDT   interp;

  if (zz<0)
    zz=-zz;
  else
    zz*=(float)z_size;
  if (zz>=z_size-1) zz=z_size-1.0001;

  for(x=0; x<x_size_pic; x++)
    for(y=0; y<y_size_pic; y++)
      {
	xx=((double)x)/im.xv;
	yy=((double)y)/im.yv;

	if (xx>=x_size-1) xx=x_size-1.0001;
	if (yy>=y_size-1) yy=y_size-1.0001;

	interp = TLI(im,xx,yy,zz);
	tmp = ( ( ((double)interp) - ((double)im.thresh2) ) * 254.0 /
		( ((double)im.thresh98) - ((double)im.thresh2) ) );
	tmp = MIN(254,MAX(0,tmp));

	if ( (compare) &&
	     (IA(im2.i,(int)(xx+0.5),(int)(yy+0.5),(int)(zz+0.5))>0) &&
	     ((x+y)%2>-10) )
	  tmp=255;

	pic[(y_size_pic-1-y)*width+x]= (unsigned char)tmp;
      }
}

/* }}} */

/* }}} */
/* {{{ write_pnm */

void write_pnm(char *fname, int width, int height, unsigned char *pic, int compare)
{
  if (!compare)
    write_pgm(fname,width,height,pic);
  else
    {
      int x, y;
      unsigned char *picr=malloc(width*height),
	*picg=malloc(width*height),
	*picb=malloc(width*height);

      for(y=0; y<height; y++)
	for(x=0; x<width; x++)
	  {
	    picr[y*width+x]=picg[y*width+x]=picb[y*width+x]=pic[y*width+x];
	    if (pic[y*width+x]==255)
	      {
		picg[y*width+x]=picb[y*width+x]=0;
		picr[y*width+x]=255;
	      }
	  }

      write_ppm(fname,width,height,picr,picg,picb);
    }
}

/* }}} */
/* {{{ main */

int main(argc, argv)
  int   argc;
  char  *argv [];
{
  /* {{{ vars */

unsigned char  *pic;
int    i, argindex=1, compare=0;
float  tmpf;
image_struct im, im2;
FDT    *edge;

/* }}} */

  if (argc<2)
    usage();

  /* {{{ read image */

avw_read(argv[argindex++],&im);

x_size = im.x;
y_size = im.y;
z_size = im.z;

/* }}} */
  /* {{{ read second (optional) image */

if ((char)argv[argindex][0]!='-')
{
  int x, y, z;
  unsigned char *orient;
  float *tmpedge;
  FDT ethresh;

  compare=1;

  avw_read(argv[argindex++],&im2);
 
  if ( (x_size!=im2.x) || (y_size!=im2.y) || (z_size!=im2.z) )
    {
      printf("Images aren't the same size - duh!\n");
      return 1;
    }

  if ( (edge=(FDT*)calloc(x_size*y_size*z_size,sizeof(FDT)))==NULL )
    {
      printf("edge calloc failed!\n");
      exit(1);
    }
  if ( (tmpedge=(float*)calloc(x_size*y_size*z_size,sizeof(float)))==NULL)
    {
      printf("tmpedge calloc failed!\n");
      exit(1);
    }
  orient=(unsigned char*)malloc(x_size*y_size*z_size);

  for(z=1; z<z_size-1; z++)
    for(y=1; y<y_size-1; y++)
      for(x=1; x<x_size-1; x++)
	{
	  double xx=IA(im2.i,x+1,y,z)-IA(im2.i,x-1,y,z),
	    yy=IA(im2.i,x,y+1,z)-IA(im2.i,x,y-1,z),
	    zz=IA(im2.i,x,y,z+1)-IA(im2.i,x,y,z-1);

	  xx*=xx; yy*=yy; zz*=zz;

	  if ( (xx>yy) && (xx>zz) )
	    IA(orient,x,y,z)=0;
	  else
	    {
	      if (yy>zz)
		IA(orient,x,y,z)=1;
	      else
		IA(orient,x,y,z)=2;
	    }

	  IA(tmpedge,x,y,z) = xx+yy+zz;
	}

  for(z=1; z<z_size-1; z++)
    for(y=1; y<y_size-1; y++)
      for(x=1; x<x_size-1; x++)
	{
	  if (IA(orient,x,y,z)==0)
	    {
	      if ( (IA(tmpedge,x,y,z)>IA(tmpedge,x-1,y,z)) && (IA(tmpedge,x,y,z)>=IA(tmpedge,x+1,y,z)) )
		IA(edge,x,y,z)=sqrt(IA(tmpedge,x,y,z))/12;
	    }
	  else
	    {
	      if (IA(orient,x,y,z)==1)
		{
		  if ( (IA(tmpedge,x,y,z)>IA(tmpedge,x,y-1,z)) && (IA(tmpedge,x,y,z)>=IA(tmpedge,x,y+1,z)) )
		    IA(edge,x,y,z)=sqrt(IA(tmpedge,x,y,z))/12;
		}
	      else
		{
		  if ( (IA(tmpedge,x,y,z)>IA(tmpedge,x,y,z-1)) && (IA(tmpedge,x,y,z)>=IA(tmpedge,x,y,z+1)) )
		    IA(edge,x,y,z)=sqrt(IA(tmpedge,x,y,z))/12;
		}
	    }
	}

  free(tmpedge);
  free(orient);

  im2.i=edge;
  im2.min=im2.max=0;
  find_thresholds(&im2,0.1);
  ethresh=((double)im2.thresh98-(double)im2.thresh2)*0.5+im2.thresh2;
  /*printf("%f %f %f\n",(float)im2.thresh2,(float)im2.thresh98,(float)ethresh);*/

  for(z=1; z<z_size-1; z++)
    for(y=1; y<y_size-1; y++)
      for(x=1; x<x_size-1; x++)
	if (IA(edge,x,y,z)<ethresh)
	  IA(edge,x,y,z)=0;
}

/* }}} */
  /* {{{ set scale etc */

tmpf = MIN(im.xv,MIN(im.yv,im.zv));

if (!strcmp(argv[argindex],"-s"))
{
  tmpf /= atof(argv[argindex+1]);
  argindex+=2;
}

im.xv/=tmpf;
im.yv/=tmpf;
im.zv/=tmpf;

x_size_pic = x_size*im.xv;
y_size_pic = y_size*im.yv;
z_size_pic = z_size*im.zv;

/* }}} */
  /* {{{ find thresholds */

im.min=im.max=0;
find_thresholds(&im,0.1);

/*printf("Thresholds: %f %f %f %f %f\n",(float)im.min,(float)im.thresh2,(float)im.thresh,(float)im.thresh98,(float)im.max);*/

/* fix thresholds if a simple binary image */
if (im.thresh2==im.thresh98)
{
  im.thresh2=im.min;
  im.thresh98=im.max;
}

if (!strcmp(argv[argindex],"-i"))
{
  im.thresh2  = atof(argv[argindex+1]);
  im.thresh98 = atof(argv[argindex+2]);
  argindex+=3;
}

/* }}} */
  /* {{{ create pic */

i = MAX(x_size_pic,MAX(y_size_pic,z_size_pic));
pic = (unsigned char *) malloc(i*i*10);

/* }}} */
  /* {{{ process slice requests */

for (i = argindex; i < argc; i++) {

  /* {{{ setup some dimensions */

int sagcorskip=0, axiskip=0,
  height=MAX(y_size_pic,z_size_pic);

if (y_size_pic>z_size_pic)
  sagcorskip=(y_size_pic-z_size_pic)/2;
else
  axiskip=(z_size_pic-y_size_pic)/2;

/* }}} */

  if (!strcmp(argv[i], "-a"))
    /* {{{ write all slices */

{
  int width=y_size_pic+2*x_size_pic;

  memset(pic,0,width*height);

  sag(0.5, im, im2, compare, pic                       + sagcorskip*width, width);
  cor(0.5, im, im2, compare, pic+y_size_pic            + sagcorskip*width, width);
  axi(0.5, im, im2, compare, pic+y_size_pic+x_size_pic + axiskip*width,    width);

  write_pnm(argv[++i],width,height,pic,compare);
}

/* }}} */
  else if (!strcmp(argv[i], "-x"))
    /* {{{ write sagittal slice */

{
  memset(pic,0,y_size_pic*height);
  sag(atof(argv[++i]), im, im2, compare, pic+sagcorskip*y_size_pic, y_size_pic);
  write_pnm(argv[++i],y_size_pic,height,pic,compare);
}

/* }}} */
  else if (!strcmp(argv[i], "-y"))
    /* {{{ write coronal slice */

{
  memset(pic,0,x_size_pic*height);
  cor(atof(argv[++i]), im, im2, compare, pic+sagcorskip*x_size_pic, x_size_pic);
  write_pnm(argv[++i],x_size_pic,height,pic,compare);
}

/* }}} */
  else if (!strcmp(argv[i], "-z"))
    /* {{{ write axial slice */

{
  memset(pic,0,x_size_pic*height);
  axi(atof(argv[++i]), im, im2, compare, pic+axiskip*x_size_pic, x_size_pic);
  write_pnm(argv[++i],x_size_pic,height,pic,compare);
}

/* }}} */

  else
    usage();
}

/* }}} */

  return(0);
}

/* }}} */
