#

#{{{ setup

#   susan.tcl - GUI for susan smoothing
#
#   Stephen Smith, FMRIB Image Analysis Group
#
#   Copyright (C) 1999-2000 University of Oxford
#
#   Part of FSL - FMRIB's Software Library
#   WWW:      http://www.fmrib.ox.ac.uk/fsl
#   Email:    fsl@fmrib.ox.ac.uk
#   
#   Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
#   Imaging of the Brain), Department of Clinical Neurology, Oxford
#   University, Oxford, UK
#   
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or (at
#   your option) any later version.
#   
#   This program is distributed in the hope that it will be useful, but in
#   order that the University as a charitable foundation protects its
#   assets for the benefit of its educational and research purposes, the
#   University makes clear that no condition is made or to be implied, nor
#   is any warranty given or to be implied, as to the accuracy of FSL, or
#   that it will be suitable for any particular purpose or for use under
#   any specific conditions.  Furthermore, the University disclaims all
#   responsibility for the use which is made of FSL.  See the GNU General
#   Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#   USA

if { [ info exists env(FSLDIR) ] } {
    set FSLDIR $env(FSLDIR)
} else {
    set FSLDIR $PXHOME/fsl
}

source ${FSLDIR}/tcl/medxstart.tcl
set auto_path [ linsert $auto_path 0 [ file dirname [ info script ] ] ]

#}}}

#{{{ proc susan

proc susan { w } {
    global susanvars
    global PXHOME
    global FSLDIR
 
    #{{{ setup main window

toplevel $w

wm title $w "SUSAN   Structure-Preserving Noise Reduction"
wm iconname $w "SUSAN"
wm iconbitmap $w @$FSLDIR/tcl/fmrib.xbm

frame $w.f

#}}}
    #{{{ setup variable defaults

set susanvars($w,dt) 3
set susanvars($w,um) 1

set thetype [ what_type 0 ]
if { $thetype > 1 && $thetype < 4 } {

    MxGetCurrentImage VolumeS
    MxGetImageOpMode $VolumeS opmodeS
    if { $opmodeS == "3D" } {
	set susanvars($w,dim) 3D
    } else {
	set susanvars($w,dim) 2D
    }

    MxQuietStatistics $VolumeS { Maximum Minimum } false Stats
    set VMin [ keylget Stats Minimum ]
    set VMax [ keylget Stats Maximum ]
    set susanvars($w,bt) [ expr ( ( $VMax - $VMin ) * 30.0 ) / 100.0 ]

} else {
    set susanvars($w,dim) 2D
    set susanvars($w,bt) 15
}

#}}}
    #{{{ top row

    frame $w.f.top

    tixOptionMenu $w.f.op -label "Dimensionality" 
    OpMenu $w.f.op [list 2D 3D]
    $w.f.op config -variable susanvars($w,dim) 

    tixControl $w.f.bt -label "Brightness threshold" -variable susanvars($w,bt) -step 1 -min 0 -selectmode immediate -options {
	entry.width 8
    }

    tixControl $w.f.dt -label "Mask SD" -variable susanvars($w,dt) -step 1 -min 0 -selectmode immediate -options {
	entry.width 5
    }

    pack $w.f.top -in $w.f -side top -padx 3 -pady 3 -expand yes -anchor w

    pack $w.f.op $w.f.bt $w.f.dt -in $w.f.top -side left -padx 5 -pady 3 -expand yes -anchor w

#}}}
    #{{{ help text

    label $w.f.label -wraplength 550 -text "SUSAN filtering works on 16 bit signed integer data.\nMask SD (approx HWHM) is in current length units (e.g. mm), not voxels.\nSet Mask SD to 0 for a (much faster) flat 3x3 voxels or 3x3x3 voxels mask." -fg "#303030"

    pack $w.f.label -in $w.f -side top -anchor s

#}}}
    #{{{ advanced options

collapsible frame $w.f.opts -title "Advanced options"

#{{{ use median?

frame $w.f.um

label $w.f.um.label -text "Use median when no neighbourhood is found "

checkbutton $w.f.um.but -variable susanvars($w,um)

pack $w.f.um.label $w.f.um.but -in $w.f.um -side left

pack $w.f.um  -in $w.f.opts.b -anchor w -pady 5

#}}}
#{{{ more USANs

set susanvars($w,maxusans) 2
set susanvars($w,usans) 0

tixControl $w.f.usans -label "Separate images to find USAN from " \
	-variable susanvars($w,usans) -step 1 -min 0 -max $susanvars($w,maxusans) -selectmode immediate -command "susan:updateusan $w"

pack $w.f.usans -in $w.f.opts.b -anchor w -pady 5

set i 1
while { $i <= $susanvars($w,maxusans) } {
    frame $w.f.usanentries($i)

    label $w.f.uel($i) -text "USAN image ${i} "
    entry $w.f.uee($i) -textvariable susanvars($w,usanentries,$i) -width 40

    set susanvars($w,ubt,$i) $susanvars($w,bt)
    tixControl $w.f.ubt$i -label "Brightness threshold" -variable susanvars($w,ubt,$i) -step 1 -min 0 -selectmode immediate -options {
	entry.width 8
    }

    pack $w.f.uel($i) $w.f.uee($i) $w.f.ubt$i -in $w.f.usanentries($i) -padx 3 -pady 3 -side left
    incr i 1
}

#}}}

pack $w.f.opts -in $w.f -side bottom -anchor w -pady 5

#}}}
    #{{{ button frame

    frame $w.btns
    frame $w.btns.b -relief raised -borderwidth 1
    
    button $w.ok -text "OK" -width 5 -command "susan:apply $w destroy"
    bind $w.ok <Return> {
        [winfo toplevel %W].ok invoke
    }
    
    button $w.apply -command "susan:apply $w keep" -text "Apply" -width 5
    bind $w.apply <Return> {
	[winfo toplevel %W].apply invoke
    }
	    
    button $w.cancel    -command "susan:destroy $w" \
	    -text "Cancel" -width 5
    bind $w.cancel <Return> {
	[winfo toplevel %W].cancel invoke
    }

    button $w.help -command "FmribWebHelp file: $FSLDIR/doc/susan/index.html" \
            -text "Help" -width 5
    bind $w.help <Return> {
        [winfo toplevel %W].help invoke
    }

    pack $w.btns.b -side bottom -fill x -padx 3 -pady 3
    pack $w.ok $w.apply $w.cancel $w.help -in $w.btns.b \
	    -side left -expand yes -padx 3 -pady 3 -fill y

#}}}
		
    pack $w.f $w.btns -expand yes -fill both
}

#}}}
#{{{ proc susan:apply

proc susan:apply { w dialog } {
    global susanvars
    global PXHOME

    foreach v { bt dt } {
	$w.f.$v update
    }

    #{{{ find and test current image

    set the_type [ what_type 1 ]
    if { $the_type < 2 } {
	return
    }

    if { $the_type > 3 } {
	MxPause "You do not have a 2D or 3D image currently selected"
	return
    }

    MxGetCurrentFolder Folder
    MxGetCurrentImage Volume

    set depth 1
    MxGetDimensions $Volume width height depth
    MxGetImageProperties $Volume pl
    set bpp [ keylget pl Bits ]
    set enc [ keylget pl Encoding ]
    if {$bpp > 16} {
	MxPause "SUSAN can only process 8 or 16 bit data"
	return
    }

#}}}
    #{{{ process USAN entries

set usanlist ""

set i 1
while { $i <= $susanvars($w,usans) } {
    set success [ MxGetPageByName $Folder $susanvars($w,usanentries,$i) usanpage($i) ]

    if { $success } {
	MxPause "Not all of the USAN images are valid"
	return
    }

    set udepth 1
    MxGetDimensions $usanpage($i) uwidth uheight udepth
    MxGetImageProperties $usanpage($i) pl
    set ubpp [ keylget pl Bits ]
    set uenc [ keylget pl Encoding ]
    MxSetCurrentPage $Volume

    if { $udepth != $depth || $uwidth != $width || $uheight != $height || $ubpp != $bpp || $uenc != $enc } {
	MxPause "Not all of the USAN images are of the same size and type as the input image"
	return
    }

    lappend usanlist $usanpage($i)
    puts "USAN $i Name=$susanvars($w,usanentries,$i) Page=$usanpage($i)"

    lappend usanlist $susanvars($w,ubt,$i)

    incr i 1
}

#}}}
    
    susan_proc $Folder $Volume $susanvars($w,bt) $susanvars($w,dt) $susanvars($w,dim) $susanvars($w,um) $susanvars($w,usans) $usanlist
    
    update idletasks
    
    if {$dialog == "destroy"} {
	susan:destroy $w
    }
}

#}}}
#{{{ proc susan:destroy

proc susan:destroy { w } {
    destroy $w
}

#}}}
#{{{ updates

proc susan:updateusan { w val } {
    global susanvars

    if { $susanvars($w,usans) == 0 } {
	pack $w.f.bt -in $w.f.top -before $w.f.dt -side left -padx 5 -pady 3 -expand yes -anchor w
    } else {
	pack forget $w.f.bt
    }

    set i 1
    while { $i <= $susanvars($w,maxusans) } {
	pack forget $w.f.usanentries($i)
	incr i 1
    }

    set i 1
    while { $i <= $susanvars($w,usans) } {
	pack $w.f.usanentries($i) -in $w.f.opts.b -anchor w -pady 5
	incr i 1
    }
}

#}}}

#{{{ tail end

susan .rename
tkwait window .rename

#}}}
